Require Import LibTactics.
Require Import ArithMisc.
Require Import Coq.ZArith.ZArith.
Local Open Scope Z_scope.
Require Import Coq.Program.Equality.
Require Import Wf.
Require Import InfiniteTypes.

(* ----------------------------------------------------------------------------------------------------------------- *)

(* Sets of relative integers. *)

(* We are interested in upward-closed subsets of [Z], so there are only three
   kinds of sets: the full set, the set of all integers including and above
   [k], and the empty set. *)

(* The following two definitions introduce the syntax of sets as well as
   their interpretation in [Z -> Prop]. *)

Inductive set : Type :=
| Full : set
| Above : Z -> set
| Void.

Definition set_denotation (s : set) (n : Z) : Prop :=
  match s with
  | Full =>
      True
  | Above k =>
      k <= n
  | Void =>
      False
  end.

(* Set inclusion. *)

Definition set_inclusion (s1 : set) (s2 : set) : Prop :=
  forall n, set_denotation s1 n -> set_denotation s2 n.

Lemma set_inclusion_reflexive:
  forall s, set_inclusion s s.
Proof.
  intros ? ?. auto.
Qed.

Hint Resolve set_inclusion_reflexive.

Lemma set_inclusion_transitive:
  forall s1 s2 s3,
  set_inclusion s1 s2 ->
  set_inclusion s2 s3 ->
  set_inclusion s1 s3.
Proof.
  unfold set_inclusion. intuition eauto.
Qed.

(* Facts about set inclusion: elimination. *)

Lemma set_inclusion_full_above:
  forall k,
  set_inclusion Full (Above k) ->
  False.
Proof.
  intros k h.
  generalize (h (k - 1)).
  simpl.
  intuition omega.
Qed.

Lemma set_inclusion_full_void:
  set_inclusion Full Void ->
  False.
Proof.
  intros h.
  generalize (h 0).
  simpl.
  intuition omega.
Qed.

Lemma set_inclusion_above_void:
  forall k,
  set_inclusion (Above k) Void ->
  False.
Proof.
  intros k h.
  generalize (h k).
  simpl.
  intuition omega.
Qed.

Lemma set_inclusion_above_elim:
  forall k1 k2,
  set_inclusion (Above k1) (Above k2) ->
  k1 >= k2.
Proof.
  intros ? ? h.
  generalize (h k1).
  unfold set_denotation.
  intuition omega.
Qed.

Hint Immediate set_inclusion_full_above set_inclusion_full_void
set_inclusion_above_void.

Hint Resolve set_inclusion_above_elim.

Lemma not_full_by_inclusion:
  forall s1 s2,
  s2 <> Full ->
  set_inclusion s1 s2 ->
  s1 <> Full.
Proof.
  intros. intro; subst.
  destruct s2; eauto.
Qed.

Lemma set_inclusion_void_elim:
  forall s,
  set_inclusion s Void ->
  s = Void.
Proof.
  intros [| |] ?. false; eauto. false; eauto. auto.
Qed.

Lemma back_to_set_inclusion:
  forall k s1 s2,
  set_denotation s1 k ->
  set_inclusion s1 s2 ->
  set_denotation s2 k.
Proof.
  intuition eauto.
Qed.

Lemma set_inclusion_antisymmetric:
  forall s1 s2,
  set_inclusion s1 s2 ->
  set_inclusion s2 s1 ->
  s1 = s2.
Proof.
  intros [ | z1 | ] [ | z2 | ] ? ?; try solve [ false; eauto ]; try congruence. 
  assert (z1 >= z2). eauto using set_inclusion_above_elim.
  assert (z2 >= z1). eauto using set_inclusion_above_elim.
  f_equal. omega.
Qed.

Hint Resolve set_inclusion_antisymmetric.

(* Facts about set inclusion: introduction. *)

Lemma set_inclusion_void:
  forall s,
  set_inclusion Void s.
Proof.
  intros ? n.
  unfold set_denotation.
  intuition auto.
Qed.

Lemma set_inclusion_full:
  forall s,
  set_inclusion s Full.
Proof.
  intros ? n.
  unfold set_denotation.
  intuition auto.
Qed.

Lemma set_inclusion_above_intro:
  forall k1 k2,
  k1 >= k2 ->
  set_inclusion (Above k1) (Above k2).
Proof.
  intros ? ? ? n.
  unfold set_denotation.
  omega.
Qed.

Hint Resolve set_inclusion_void set_inclusion_full
set_inclusion_above_intro.

(* Translation of a subset of [Z]. *)

Definition translate (i : Z) (s : set) : set :=
  match s with
  | Full =>
      Full
  | Above k =>
      Above (k + i)
  | Void =>
      Void
  end.

Lemma denotation_translate:
  forall i s n,
  set_denotation (translate i s) n <->
  set_denotation s (n - i).
Proof.
  unfold set_denotation.
  intros ? ? ?.
  dependent destruction s; simpl; intuition omega.
Qed.

Lemma translate_monotonic:
  forall i s1 s2,
  set_inclusion s1 s2 ->
  set_inclusion (translate i s1) (translate i s2).
Proof.
  intros ? ? ? ? ?.
  repeat (rewrite denotation_translate).
  eauto.
Qed.

Hint Resolve translate_monotonic.

(* ----------------------------------------------------------------------------------------------------------------- *)

(* The subtyping algorithms. *)

(* The function [combine] is a key component of the subtyping algorithms. If
   [s1] and [s2] respectively represent answers to subtyping queries on the
   domain and codomain, then [combine s1 s2] represents the answer to a
   subtyping query on a pair of arrows. *)

Definition combine (s1 s2 : set) : set :=
  match s1, s2 with
  | Void, _
  | _, Void =>
      Void
  | Above k1, Above k2 =>
      match k2 ?= -k1 with
      | Gt =>
          (* k2 > -k1 *)
          Void
      | Lt
      | Eq =>
          (* k2 <= -k1 *)
          Above k2
      end
  | _, s2 =>
      s2
  end.

(* The subtyping algorithm is defined in two versions: one accepts a pair of
   types as input, while the other accepts a pair of skeletons. Both return
   a set. *)

(* The subtyping algorithm is defined as the greatest fixed point of the
   following higher-order functions. *)

Definition subtype (self : ty -> ty -> set) (ty1 : ty) (ty2 : ty) : set :=
  match ty1, ty2 with
  | TyBullet ty1, _ =>
      translate 1 (self ty1 ty2)
  | _, TyBullet ty2 =>
      translate (-1) (self ty1 ty2)
  | TyLeaf, TyLeaf =>
      Above 0
  | TyArrow ty11 ty12, TyArrow ty21 ty22 =>
      combine (self ty21 ty11) (self ty12 ty22)
  | TyLeaf, TyArrow _ _
  | TyArrow _ _, TyLeaf =>
      Void
  end.

Definition subskeleton (self : skeleton -> skeleton -> set) (s1 : skeleton) (s2 : skeleton) : set :=
  match s1, s2 with
  | SkLeaf i1, SkLeaf i2 =>
      Above (Z_of_nat i1 - Z_of_nat i2)
  | SkArrow s11 s12, SkArrow s21 s22 =>
      combine (self s21 s11) (self s12 s22)
  | SkLeaf _, SkArrow _ _
  | SkArrow _ _, SkLeaf _ =>
      Void
  end.

(* The initial approximation. *)

Definition top A (_ _ : A) : set :=
  Full.

Implicit Arguments top [ A ].

(* ----------------------------------------------------------------------------------------------------------------- *)

(* We check that the above higher-order functions are monotonic with
   respect to set inclusion. This implies that these functions have
   a unique, well-defined, greatest fixed point. *)

Definition pointwise_inclusion (A : Type) (f1 f2 : A -> A -> set) : Prop :=
  forall x1 x2 : A, set_inclusion (f1 x1 x2) (f2 x1 x2).

Implicit Arguments pointwise_inclusion [ A ].

Lemma pointwise_inclusion_reflexive:
  forall A (f : A -> A -> set),
  pointwise_inclusion f f.
Proof.
  intros. intro. intuition eauto.
Qed.

Lemma pointwise_inclusion_transitive:
  forall A (f1 f2 f3 : A -> A -> set),
  pointwise_inclusion f1 f2 ->
  pointwise_inclusion f2 f3 ->
  pointwise_inclusion f1 f3.
Proof.
  unfold pointwise_inclusion. intuition eauto using set_inclusion_transitive.
Qed.

Hint Resolve pointwise_inclusion_reflexive.

(* [combine] is monotonic -- a slightly non-obvious property. *)

Lemma combine_monotonic:
  forall s1 t1 s2 t2,
  set_inclusion s1 t1 ->
  set_inclusion s2 t2 ->
  set_inclusion (combine s1 s2) (combine t1 t2).
Proof.
  intros.
  (* Examine 89 cases. *)
  destruct s1 as [| j1 |]; destruct s2 as [| j2 |]; destruct t1 as [| k1 |]; destruct t2 as [| k2 |];
  eauto; try solve [ false; eauto ]; simpl.
  (* Only two cases remain. The first one is trivial: *)
  destruct (j2 ?= -j1); eauto.
  (* The second one requires some more work: *)
  assert (j1 >= k1). eauto.
  assert (j2 >= k2). eauto.
  generalize (Zgt_compare j2 (-j1)).
  generalize (Zle_compare k2 (-k1)).
  destruct (j2 ?= -j1); destruct (k2 ?= -k1); eauto; try solve [ intros; false; omega ].
Qed.

(* The following corollaries are immediate. *)

Lemma subtype_monotonic:
  forall self1 self2 : ty -> ty -> set,
  pointwise_inclusion self1 self2 ->
  pointwise_inclusion (subtype self1) (subtype self2).
Proof.
  intros ? ? h ty1 ty2.
  destruct ty1 as [| ty11 ty12 |]; destruct ty2 as [| ty21 ty22 |]; simpl; eauto.
  eauto using combine_monotonic.
Qed.

Lemma subskeleton_monotonic:
  forall self1 self2 : skeleton -> skeleton -> set,
  pointwise_inclusion self1 self2 ->
  pointwise_inclusion (subskeleton self1) (subskeleton self2).
Proof.
  intros ? ? h s1 s2.
  destruct s1 as [| s11 s12 ]; destruct s2 as [| s21 s22 ]; simpl; eauto.
  eauto using combine_monotonic.
Qed.

(* ----------------------------------------------------------------------------------------------------------------- *)

(* We now prove that the subtyping algorithm cannot produce the value
   [Full]. In other words, this value is useful only because it plays the role
   of the greatest element of the lattice [set] -- it is used to initialize
   the iterative computation of the greatest fixed point. *)

(* We begin with a few technical preliminaries. *)

Lemma self_not_full_strengthen:
  forall A : Type,
  forall sub : (A -> A -> set) -> A -> A -> set,
  forall self,
  pointwise_inclusion self (sub self) ->
  forall x1 x2,
  sub self x1 x2 <> Full ->
  self x1 x2 <> Full.
Proof.
  eauto using not_full_by_inclusion.
Qed.

Implicit Arguments self_not_full_strengthen [ A sub self ].

Lemma translate_not_full:
  forall i s,
  s <> Full ->
  translate i s <> Full.
Proof.
  intros. destruct s; simpl; try discriminate; auto.
Qed.

Hint Resolve translate_not_full.

(* An obvious but important property of [combine] is that [combine s1 s2]
   can be [Full] only if [s2] is [Full]. *)

Lemma combine_not_full:
  forall s1 s2,
  s2 <> Full ->
  combine s1 s2 <> Full.
Proof.
  destruct s1; destruct s2; intro; auto; try discriminate.
  simpl. destr_zcompare; discriminate.
Qed.

(* As a consequence of the above property, tail-finiteness is a sufficient
   condition to guarantee that the algorithm does not produce [Full]. This
   fact holds for any pre-fixpoint. *)

Lemma subtype_not_full:
  forall self,
  pointwise_inclusion self (subtype self) ->
  forall ty1, tf ty1 ->
  forall ty2, tf ty2 ->
  self ty1 ty2 <> Full.
Proof.
  introv prefixpoint.
  induction 1.
  (* Leaf *)
  induction 1; apply (self_not_full_strengthen prefixpoint); simpl; try discriminate; eauto.
  (* Arrow *)
  induction 1; apply (self_not_full_strengthen prefixpoint); simpl; try discriminate; eauto.
    (* Arrow Arrow *)
    eauto using combine_not_full.
  (* Bullet *)
  intros. apply (self_not_full_strengthen prefixpoint); simpl. eauto.
Qed.

(* ----------------------------------------------------------------------------------------------------------------- *)

(* We now prove that the algorithm is sound. *)

(* A technical lemma, which is really just a tautology. *)

Lemma auxiliary_analysis:
  forall a1 a2 k,
  set_denotation (
    match a2 ?= - a1 with
    | Eq => Above a2
    | Lt => Above a2
    | Gt => Void
    end
  ) k ->
  a2 <= -a1 /\ a2 <= k.
Proof.
  intros ? ? ?.
  destr_zcompare; simpl; intro; omega.
Qed.

(* The soundness statement is proved directly. We show that any pre-fixpoint
   is sound. The statement is somewhat convoluted, because [sksub] involves
   two natural integers, whereas our sets are subsets of [Z]. If the
   difference [k2 - k1] is a member of the set produced by the algorithm, [set
   s1 s2], then the skeletons [s1] and [s2] are related, that is, [sksub k1 s1
   k2 s2] holds. *)

(* TEMPORARY probleme: je croyais naivement que la *suite de fonctions* selfsk
   va atteindre un jour son point fixe, ce qui risque d'etre faux -- la
   convergence ne se fait que pour une requete fixee, et le temps de
   convergence depend de la requete. Cet enonce reste utilisable, j'espere,
   pour etablir le corollaire final, mais la construction de la fonction
   limite sera plus difficile. *)

(* TEMPORARY en fait, la limite de cette suite de fonctions existe -- on
   doit pouvoir la construire -- mais n'est pas un element de la suite...
   faut-il construire cette limite et prouver qu'elle est sound? il
   devrait suffire de montrer qu'elle est bien un pre-fixpoint. La
   completude serait plus dure? *)

(* TEMPORARY ou bien faut-il laisser tomber subskeleton et tenter de
   prouver directement la soundness de subtype? en tout cas, on aura
   le meme probleme de limite. *)

Theorem subskeleton_sound:
  forall self,
  pointwise_inclusion self (subskeleton self) ->
  forall s1 s2 k1 k2,
  set_denotation (self s1 s2) (Z_of_nat k2 - Z_of_nat k1) ->
  sksub k1 s1 k2 s2.
Proof.
  intros ? prefixpoint.
  cofix.
  introv h.
  generalize (prefixpoint _ _ _ h). clear h. intro h.
  dependent destruction s1; dependent destruction s2; simpl in h; try solve [ false; auto ].
  (* Leaf Leaf *)
  constructor. omega.
  (* Arrow Arrow *)
  generalize h. clear h.
  generalize (refl_equal (self s2_1 s1_1)); pattern (self s2_1 s1_1) at -1; destruct (self s2_1 s1_1) as [| a1 |];
  generalize (refl_equal (self s1_2 s2_2)); pattern (self s1_2 s2_2) at -1; destruct (self s1_2 s2_2) as [| a2 |];
  simpl; intros h2 h1 h;
  try solve [ false; auto ].
  (* The first three cases involve a call to [self] that returns
     [Full]. These cases could be eliminated if one were willing
     to make the hypothesis that the skeletons are finite. *)
  apply (SkSubArrow 0); apply subskeleton_sound; [ rewrite h1 | rewrite h2 ]; auto.
  apply (SkSubArrow 0); rewrite arith_plus_zero; apply subskeleton_sound; [ rewrite h1 | rewrite h2 ]; simpl; auto.
  apply (SkSubArrow (Zabs_nat (a1 + Z_of_nat k2 - Z_of_nat k1)));
  apply subskeleton_sound; [ rewrite h1 | rewrite h2 ]; simpl; auto.
    rewrite inj_plus.
    generalize (Z_nat_Z_abs_either_way (a1 + Z_of_nat k2 - Z_of_nat k1)).
    omega.
  (* The last case is the really interesting one. Here, we assume
     that the calls to [self] have returned [Above a1] and [Above a2].
     An auxiliary analysis shows that this implies [a2 <= -a1] and
     [a2 <= k2 - k1]. *)
  generalize (auxiliary_analysis _ _ _ h); intros [ ? ? ]. clear h.
  set (slack := Z_of_nat k2 - Z_of_nat k1 - a2).
  assert (slack >= 0). unfold slack. omega.
  assert (Z_of_nat k2 - Z_of_nat ((k1 + Zabs_nat slack)%nat) = a2).
    rewrite inj_plus.
    rewrite Z_nat_Z_abs; auto.
    unfold slack.
    omega.
  apply (SkSubArrow (Zabs_nat slack));
    apply subskeleton_sound; [ rewrite h1 | rewrite h2 ]; simpl;
    omega.
Qed.

Lemma subskeleton_sound_corollary:
  forall self,
  pointwise_inclusion self (subskeleton self) ->
  forall s1 s2,
  set_denotation (self s1 s2) 0 ->
  sksub 0 s1 0 s2.
Proof.
  intros. eauto using subskeleton_sound.
Qed.

(* ----------------------------------------------------------------------------------------------------------------- *)

(* We now prove that the algorithm is complete. *)

(* Completeness is slightly trickier than soundness, because it does not hold
   of any post-fixpoint, but only of the successive approximations that lead
   from Top down to the greatest fixed point. *)

(* An approximation is complete if it contains [sksub]. *)

Definition complete (self : skeleton -> skeleton -> set) :=
  forall s1 s2 k1 k2, sksub k1 s1 k2 s2 -> set_denotation (self s1 s2) (Z_of_nat k2 - Z_of_nat k1).

(* The main lemma states that if, some approximation is complete, then the
   next approximation in the sequence is complete as well. *)

Lemma subskeleton_complete_preliminary:
  forall self,
  complete self ->
  complete (subskeleton self).
Proof.
  introv h ss.
  dependent destruction s1; dependent destruction s2; dependent destruction ss; simpl.
  (* Leaf Leaf *)
  omega.
  (* Arrow Arrow *)
  generalize (h _ _ _ _ ss1).
  generalize (h _ _ _ _ ss2).
  clear h ss1 ss2.
  generalize (self s2_1 s1_1); intro s1.
  generalize (self s1_2 s2_2); intro s2.
  (* We have cleaned up, now let's prove this by case analysis. *)
  dependent destruction s1; dependent destruction s2; simpl; try solve [ intuition eauto ];
  repeat (rewrite inj_plus); intros.
  omega.
  destr_zcompare; simpl; omega.
Qed.

(* Since the initial approximation, [top], is obviously complete,
   there follows that the entire sequence consists of complete approximations.
   The greatest fixed point, which is the limit of the sequence, is complete
   as well. *)

(* TEMPORARY careful: the limit is not an element of the sequence! *)

Theorem subskeleton_complete:
  always (now complete) (orbit subskeleton (@top skeleton)).
Proof.
  assert (h : complete (@top skeleton)).
    intros ? ? ? ? ?. simpl. auto.
  generalize (@top skeleton) h.
  cofix.
  intros self ?.
  rewrite (seq_decompose (orbit _ _)); simpl. constructor.
  eauto using subskeleton_complete_preliminary.
  eauto.
Qed.

(* ----------------------------------------------------------------------------------------------------------------- *)

(* We now prove that the algorithm converges, that is, the sequence of
   approximations is eventually constant. This is a non-obvious fact, because
   our lattice of subsets of [Z] has infinite height. The proof, however, is
   not difficult. It exploits the hypothesis that types are tail-finite. *)

(* A sequence is monotonic if the next element is a subset of the previous
   element. *)

Definition monotonic :=
  always (rel set_inclusion).

(* Some properties of monotonicity. *)

Lemma monotonic_tail:
  forall s a,
  monotonic (SeqCons s a) ->
  monotonic s.
Proof.
  unfold monotonic. eauto using A_cons_tail.
Qed.

Hint Resolve monotonic_tail.

Lemma monotonic_inclusion:
  forall s b a,
  monotonic (SeqCons (SeqCons s b) a) ->
  set_inclusion b a.
Proof.
  introv h. generalize (A_cons_head h). unfold rel. auto.
Qed.

(* A sequence converges if it is eventually constant. *)

Definition converges (s : sequence set) :=
  exists limit, eventually (always (now (= limit))) s.

(* Some properties of convergence. *)

Lemma converges_compatible:
  compatible converges.
Proof.
  intros s1 s2 [ limit h1 ] h2. exists limit.
  assert (compatible (eventually (always (now (= limit))))).
    eauto using now_compatible, always_compatible, eventually_compatible.
  eauto.
Qed.

Lemma converges_tl:
  forall s,
  converges (tl s) ->
  converges s.
Proof.
  intros [ s x ]. simpl. intros [ limit ? ]. exists limit. apply EventuallyLater. auto.
Qed.

Lemma converges_translate:
  forall i s,
  converges s ->
  converges (map (translate i) s).
Proof.
  intros i s [ limit h ]. exists (translate i limit).
  induction h.
  apply EventuallyNow.
    generalize s H. clear s H. cofix. intros [ s x ] h. dependent destruction h.
    rewrite (seq_decompose (map _ _)); simpl. constructor. eauto. unfold now. unfold now in H. subst. auto.
  rewrite (seq_decompose (map _ _)); simpl. apply EventuallyLater. eauto.
Qed.

(* The following is an obvious, but important, property of the [combine]
   function. This property was used above already when we proved that the
   greatest fixed point is never [Full]. *)

Lemma combine_characteristic:
  forall s1 s2,
  combine s1 s2 = Void \/ combine s1 s2 = s2.
Proof.
  intros [| k1 |] [| k2 |]; simpl; intuition auto.
  destr_zcompare; intuition auto.
Qed.

Definition not_void :=
  now (fun s => s <> Void).

(* Now, consider two sequences [s1] and [s2], as well their pointwise
   combination, [zip combine s1 s2]. Because of the above property,
   if the combination is never [Void], then it always [s2]. *)

Lemma zip_combine_characteristic:
  forall s1 s2,
  always not_void (zip combine s1 s2) ->
  bisimilar (zip combine s1 s2) s2.
Proof.
  cofix.
  intros [ s1 x1 ] [ s2 x2 ].
  rewrite (seq_decompose (zip combine (SeqCons s1 x1) (SeqCons s2 x2))). simpl. intro h.
  dependent destruction h. constructor. eauto. generalize (combine_characteristic x1 x2).
  unfold not_void, now in H. intuition auto.
Qed.

(* As an immediate corollary, if the combination of [s1] and [s2] is never
   void, and if furthermore [s2] converges, then the combination converges
   as well, since it is just [s2]. *)

Lemma zip_combine_not_void_converges:
  forall s1 s2,
  always not_void (zip combine s1 s2) ->
  converges s2 ->
  converges (zip combine s1 s2).
Proof.
  intros.
  assert (bisimilar (zip combine s1 s2) s2).
    eauto using zip_combine_characteristic.
  eauto using converges_compatible.
Qed.

(* On the other hand, the combination of [s1] and [s2] is a monotonic
   sequence... *)

Lemma zip_combine_monotonic:
  forall s1 s2,
  monotonic s1 ->
  monotonic s2 ->
  monotonic (zip combine s1 s2).
Proof.
  eapply zip_monotonic. eauto using combine_monotonic.
Qed.

(* ... so, if it is ever [Void], then it must remain [Void] forever... *)

Lemma void_now_void_forever:
  forall s,
  monotonic s ->
  now (= Void) s ->
  always (now (= Void)) s.
Proof.
  cofix. intros [[ s y ] x ] m v.
  unfold now in v. subst.
  generalize (monotonic_inclusion _ _ _ m). intro v.
  generalize (set_inclusion_void_elim _ v). intro. subst.
  constructor.
    apply void_now_void_forever. eauto. eauto.
    eauto.
Qed.

Lemma void_eventually_void_forever:
  forall s,
  monotonic s ->
  eventually (now (= Void)) s ->
  eventually (always (now (= Void))) s.
Proof.
  induction 2.
  apply EventuallyNow. eauto using void_now_void_forever.
  apply EventuallyLater. eauto.
Qed.

(* ... which implies that it converges. *)

Lemma void_sticks:
  forall s,
  monotonic s ->
  ~ always not_void s ->
  converges s.
Proof.
  intros.
  assert (eventually (now (= Void)) s).
    assert (h : neg (always not_void) s). auto.
    eapply to_ss. eexact h. rewrite NA. apply E_covariant. intros [ ? ? ]. unfold neg, not_void, now. apply NNPP.
  exists Void.
  eauto using void_eventually_void_forever.
Qed.

(* In summary, either the sequence [zip combine s1 s2] hits [Void] at some
   point, in which case it remains [Void] forever, so it converges; or it
   never hits [Void], in which case it is equal to [s2], which by hypothesis
   converges. In either case, [zip combine s1 s2] converges. *)

Lemma zip_combine_converges:
  forall s2,
  monotonic s2 ->
  converges s2 ->
  forall s1,
  monotonic s1 ->
  converges (zip combine s1 s2).
Proof.
  introv m2 c2 m1.
  destruct (classic (always not_void (zip combine s1 s2))).
    eauto using zip_combine_not_void_converges.
    assert (monotonic (zip combine s1 s2)).
      eauto using zip_combine_monotonic.
    eauto using void_sticks.
Qed.

(* It would be possible to state an even stronger property of [combine]:
   to prove that the sequence [zip combine s1 s2] converges, it should
   suffice to assume that the sequence [s1] is not constantly [Full].
   Indeed, under this assumption, neither [s1] nor [s2] are constantly
   [Full] -- or the combination converges, and we are done. Neither are
   constantly [Void], or we are done again. So both [s1] and [s2] must
   eventually evolve in the image of [Above]. The test [k2 ?= -k1] must
   at some point yield [Gt], because both [k1] and [k2] increase. But
   at this point, the combination becomes [Void], and we are done. I
   have not formally checked this property, because it is not required
   in the proof that follows. *)

(* We now consider the sequence of sets obtained as successive approximate
   answers to a fixed subtyping query. *)

Definition sample A (ty1 ty2 : A) : sequence (A -> A -> set) -> sequence set :=
  map (fun f => f ty1 ty2).

Implicit Arguments sample [ A ].

Definition approximations (ty1 ty2 : ty) : sequence set :=
  sample ty1 ty2 (orbit subtype (@top ty)).

(* We check, by cases on the structure of the types [ty1] and [ty2], that this
   sequence of approximations converges, provided a sequence that is obtained
   one level down in the type structure converges. The only non-trivial case
   is that of the arrow, where the above property of [combine] is
   exploited. *)

Lemma approximations_leaf_leaf:
  converges (approximations TyLeaf TyLeaf).
Proof.
  exists (Above 0).
  unfold approximations. rewrite (seq_decompose (sample _ _ _)); simpl.
  apply EventuallyLater. apply EventuallyNow. generalize (@top ty). cofix.
  intro self. rewrite (seq_decompose (map _ _)); simpl. constructor. eauto. eauto.
Qed.

Lemma approximations_leaf_arrow:
  forall ty1 ty2,
  converges (approximations TyLeaf (TyArrow ty1 ty2)).
Proof.
  exists Void.
  unfold approximations. rewrite (seq_decompose (sample _ _ _)); simpl.
  apply EventuallyLater. apply EventuallyNow. generalize (@top ty). cofix.
  intro self. rewrite (seq_decompose (map _ _)); simpl. constructor. eauto. eauto.
Qed.

Lemma approximations_arrow_leaf:
  forall ty1 ty2,
  converges (approximations (TyArrow ty1 ty2) TyLeaf).
Proof.
  exists Void.
  unfold approximations. rewrite (seq_decompose (sample _ _ _)); simpl.
  apply EventuallyLater. apply EventuallyNow. generalize (@top ty). cofix.
  intro self. rewrite (seq_decompose (map _ _)); simpl. constructor. eauto. eauto.
Qed.

Lemma approximations_bullet_left_preliminary:
  forall ty1 ty2,
  forall self,
  bisimilar
    (sample (TyBullet ty1) ty2 (orbit subtype (subtype self)))
    (map (translate 1) (sample ty1 ty2 (orbit subtype self))).
Proof.
  intros ? ?. cofix. intro.
  rewrite (seq_decompose (sample _ _ (orbit _ _))); simpl.
  rewrite (seq_decompose (map _ (sample _ _ _))); simpl.
  constructor.
    apply approximations_bullet_left_preliminary. auto.
Qed.

Lemma approximations_bullet_right_preliminary:
  forall ty1 ty2,
  nab ty1 ->
  forall self,
  bisimilar
    (sample ty1 (TyBullet ty2) (orbit subtype (subtype self)))
    (map (translate (-1)) (sample ty1 ty2 (orbit subtype self))).
Proof.
  intros ? ? n. 
  destruct ty1; dependent destruction n.
  cofix. intro.
  rewrite (seq_decompose (sample _ _ (orbit _ _))); simpl.
  rewrite (seq_decompose (map _ (sample _ _ _))); simpl.
  constructor.
    apply approximations_bullet_right_preliminary. auto.
  cofix. intro.
  rewrite (seq_decompose (sample _ _ (orbit _ _))); simpl.
  rewrite (seq_decompose (map _ (sample _ _ _))); simpl.
  constructor.
    apply approximations_bullet_right_preliminary. auto.
Qed.

Lemma approximations_bullet_left_characterization:
  forall ty1 ty2,
  bisimilar (tl (approximations (TyBullet ty1) ty2)) (map (translate 1) (approximations ty1 ty2)).
Proof.
  intros. simpl. apply approximations_bullet_left_preliminary.
Qed.

Lemma approximations_bullet_right_characterization:
  forall ty1 ty2,
  nab ty1 ->
  bisimilar (tl (approximations ty1 (TyBullet ty2))) (map (translate (-1)) (approximations ty1 ty2)).
Proof.
  intros. simpl. apply approximations_bullet_right_preliminary. auto.
Qed.

Lemma approximations_bullet_left:
  forall ty1 ty2,
  converges (approximations ty1 ty2) ->
  converges (approximations (TyBullet ty1) ty2).
Proof.
  intros. apply converges_tl. eapply converges_compatible. eapply converges_translate. eassumption.
  eauto using approximations_bullet_left_characterization.
Qed.

Lemma approximations_bullet_right:
  forall ty1 ty2,
  nab ty1 ->
  converges (approximations ty1 ty2) ->
  converges (approximations ty1 (TyBullet ty2)).
Proof.
  intros. apply converges_tl. eapply converges_compatible. eapply converges_translate. eassumption.
  eapply bisimilar_symmetric. eauto using approximations_bullet_right_characterization.
Qed.

Lemma approximations_arrow_characterization_preliminary:
  forall ty11 ty12 ty21 ty22,
  forall self,
  bisimilar
    (sample (TyArrow ty11 ty12) (TyArrow ty21 ty22) (orbit subtype (subtype self)))
    (zip combine (sample ty21 ty11 (orbit subtype self)) (sample ty12 ty22 (orbit subtype self))).
Proof.
  intros ? ? ? ?. cofix. intro.
  rewrite (seq_decompose (sample _ _ (orbit _ _))); simpl.
  rewrite (seq_decompose (zip _ _ _)); simpl.
  constructor.
    apply approximations_arrow_characterization_preliminary. auto.
Qed.

Lemma approximations_arrow_characterization:
  forall ty11 ty12 ty21 ty22,
  bisimilar
    (tl (approximations (TyArrow ty11 ty12) (TyArrow ty21 ty22)))
    (zip combine (approximations ty21 ty11) (approximations ty12 ty22)).
Proof.
  intros. simpl. apply approximations_arrow_characterization_preliminary.
Qed.

Lemma approximations_monotonic_preliminary:
  forall ty1 ty2,
  forall self,
  pointwise_inclusion (subtype self) self ->
  monotonic (sample ty1 ty2 (orbit subtype self)).
Proof.
  unfold monotonic, sample.
  intros ? ?. cofix. intro.
  rewrite (seq_decompose (map _ _)); simpl.
  constructor.
  eapply approximations_monotonic_preliminary. eauto using subtype_monotonic.
  rewrite (seq_decompose (map _ _)); simpl.
  eauto.
Qed.

Lemma approximations_monotonic:
  forall ty1 ty2,
  monotonic (approximations ty1 ty2).
Proof.
  unfold approximations. intros.
  apply approximations_monotonic_preliminary.
  intros ? ?. unfold top. eauto.
Qed.

Hint Resolve approximations_monotonic.

Lemma approximations_arrow:
  forall ty11 ty12 ty21 ty22,
  converges (approximations ty12 ty22) ->
  converges (approximations (TyArrow ty11 ty12) (TyArrow ty21 ty22)).
Proof.
  intros. 
  assert (c : converges (zip combine (approximations ty21 ty11) (approximations ty12 ty22))).
    apply zip_combine_converges; eauto.
  apply converges_tl. eapply converges_compatible. eexact c.
  eauto using approximations_arrow_characterization.
Qed.

Hint Resolve approximations_leaf_leaf approximations_leaf_arrow approximations_arrow_leaf
approximations_arrow approximations_bullet_left approximations_bullet_right.

(* Now that each case has been checked, there only remains to put everything
   together, by induction over tail-finiteness proofs. *)

Lemma approximations_converges:
  forall ty1,
  tf ty1 ->
  forall ty2,
  tf ty2 ->
  converges (approximations ty1 ty2).
Proof.
  induction 1; induction 1; eauto. (* a swift proof! *)
Qed.

(* ----------------------------------------------------------------------------------------------------------------- *)

(* We now examine how the algorithms behave with respect to shifting. *)

(* An approximation of the subtyping algorithm over skeletons is natural if it
   behaves in the expected way with respect to shifting. *)

Definition natural (self : skeleton -> skeleton -> set) :=
  forall k1 s1 k2 s2,
  self (shift k1 s1) (shift k2 s2) = translate (Z_of_nat k1 - Z_of_nat k2) (self s1 s2).

(* The initial approximation is natural. *)

Lemma natural_top:
  natural (@top skeleton).
Proof.
  intros ? ? ? ?. auto.
Qed. 

(* If some approximation is natural, then the next approximation is natural as
   well. This follows directly from a property of [combine]. *)

Lemma combine_translate:
  forall k1 k2 s1 s2,
  k1 + k2 = 0 ->
  combine (translate k1 s1) (translate k2 s2) = translate k2 (combine s1 s2).
Proof.
  intros ? ? [| a1 |] [| a2 |] ?; simpl; auto.
  replace (a2 + k2 ?= - (a1 + k1)) with (a2 ?= -a1).
  destr_zcompare; auto.
  destr_zcompare; destr_zcompare; auto; false; omega.
Qed.

Lemma natural_subskeleton:
  forall self,
  natural self ->
  natural (subskeleton self).
Proof.
  intros ? h ? ? ? ?.
  dependent destruction s1; dependent destruction s2; simpl; auto.
  f_equal. repeat (rewrite inj_plus). omega.
  repeat (rewrite h). apply combine_translate. omega.
Qed.

(* ----------------------------------------------------------------------------------------------------------------- *)

(* We now examine how the subtyping algorithm behaves with respect to
   extensional equality of skeletons. This is trivial. *)

(* An approximation of the subtyping algorithm over skeletons is nice if it
   is compatible with [skeq]. *)

Definition nice (self : skeleton -> skeleton -> set) :=
  forall s1 s2 t1 t2,
  skeq s1 t1 ->
  skeq s2 t2 ->
  self s1 s2 = self t1 t2.

(* The initial approximation is nice. *)

Lemma nice_top:
  nice (@top skeleton).
Proof.
  intros ? ? ? ? ? ?. auto.
Qed.

(* If some approximation is nice, then the next approximation is nice as
   well. *)

Lemma nice_subskeleton:
  forall self,
  nice self ->
  nice (subskeleton self).
Proof.
  intros ? h ? ? ? ? h1 h2.
  dependent destruction s1; dependent destruction s2;
  dependent destruction h1; dependent destruction h2;
  simpl; auto.
  f_equal; eauto.
Qed.

(* ----------------------------------------------------------------------------------------------------------------- *)

(* We now prove that the two variants of the algorithm -- the one that
   works with types, and the one that works with skeletons -- coincide. *)

(* TEMPORARY nettoyer ce qui suit... *)

(* The type-based algorithm is complete with respect to the skeleton-based
   algorithm if it produces larger sets. *)

Definition rcomplete (selfty : ty -> ty -> set) (selfsk : skeleton -> skeleton -> set) : Prop :=
  forall ty1 ty2 fty1 fty2,
  set_inclusion
    (selfsk (sk 0 ty1 fty1) (sk 0 ty2 fty2))
    (selfty ty1 ty2).

(* This auxiliary lemma is isolated in order to avoid duplicating the case
   of a bullet on the right-hand side. *)

Lemma rcompleteness_bullet_right:
  forall selfty selfsk,
  rcomplete selfty selfsk ->
  natural selfsk ->
  nice selfsk ->
  pointwise_inclusion (subskeleton selfsk) selfsk ->
  forall ty1 ty2 fty1 fty2,
  set_inclusion
    (subskeleton selfsk (sk 0 ty1 fty1) (sk 0 (TyBullet ty2) fty2))
    (translate (-1) (selfty ty1 ty2)).
Proof.
  intros ? ? h na ni postfixpoint ? ? ? ? k hk.
  rewrite sk_bullet in hk.
  rewrite (nice_subskeleton _ ni) 
    with (t1 := shift 0 (sk 0 ty1 fty1))
         (t2 := shift 1 (sk 0 ty2 (finite_hereditary_bullet (TyBullet ty2) fty2 ty2 refl)))
  in hk; [| eauto using shift_zero | eauto using sk_shift_zero ].
  rewrite (natural_subskeleton _ na) in hk.
  replace (Z_of_nat 0 - Z_of_nat 1) with (-1) in hk; try solve [ simpl; auto ].
  eapply back_to_set_inclusion. eassumption.
  apply translate_monotonic;
  eauto using set_inclusion_transitive. (* postfixpoint and h are used *)
Qed.

(* This basic completeness lemma states that: if, at some iteration, the
   type-based approximation is complete with respect to the skeleton-based
   approximation, then, at the next iteration, this property still holds.  We
   use the (somewhat trivial) facts that the skeleton-based approximation is
   natural and nice. We also use the fact that it is a post-fixpoint: this is
   necessary because the type-based algorithm converges more slowly -- it is
   slowed down by the extra bullets -- so, at a given iteration, it produces
   larger sets than the skeleton-based algorithm. *)

Lemma basic_rcompleteness:
  forall selfty selfsk,
  rcomplete selfty selfsk ->
  natural selfsk ->
  nice selfsk ->
  pointwise_inclusion (subskeleton selfsk) selfsk ->
  rcomplete (subtype selfty) (subskeleton selfsk).
Proof.
  intros ? ? h na ni postfixpoint ? ? ? ? k hk.
  dependent destruction ty1.
  (* Leaf *)
    dependent destruction ty2; simpl.
    (* Leaf Leaf *)
    rewrite sk_leaf in hk.
    rewrite sk_leaf in hk. simpl in hk. auto.
    (* Leaf Arrow *)
    rewrite sk_leaf in hk.
    rewrite sk_arrow in hk. simpl in hk. auto.
    (* Leaf Bullet *)
    eapply back_to_set_inclusion. eassumption.
    eapply rcompleteness_bullet_right; eauto.
  (* Arrow *)
    dependent destruction ty2; simpl.
    (* Arrow Leaf *)
    rewrite sk_arrow in hk.
    rewrite sk_leaf in hk. simpl in hk. auto.
    (* Arrow Arrow *)
    rewrite sk_arrow in hk.
    rewrite sk_arrow in hk. simpl in hk.
    eapply back_to_set_inclusion. eassumption.
    apply combine_monotonic; eauto.
    (* Arrow Bullet *)
    eapply back_to_set_inclusion. eassumption.
    eapply rcompleteness_bullet_right; eauto.
  (* Bullet *)
  rewrite sk_bullet in hk.
  rewrite (nice_subskeleton _ ni)
    with (t1 := shift 1 (sk 0 ty1 (finite_hereditary_bullet (TyBullet ty1) fty1 ty1 refl)))
         (t2 := shift 0 (sk 0 ty2 fty2))
  in hk; [| eauto using sk_shift_zero | eauto using shift_zero ].
  rewrite (natural_subskeleton _ na) in hk.
  replace (Z_of_nat 1 - Z_of_nat 0) with 1 in hk; try solve [ simpl; auto ].
  eapply back_to_set_inclusion. eassumption.
  apply translate_monotonic;
  eauto using set_inclusion_transitive. (* postfixpoint and h are used *)
Qed.

(* The type-based algorithm is sound with respect to the skeleton-based
   algorithm if it produces smaller sets. *)

Definition rsound (selfty : ty -> ty -> set) (selfsk : skeleton -> skeleton -> set) : Prop :=
  forall ty1 ty2 fty1 fty2,
  set_inclusion
    (selfty ty1 ty2)
    (selfsk (sk 0 ty1 fty1) (sk 0 ty2 fty2)).

(* This basic soundness lemma states that: if, at some iteration, the
   type-based fixed point is sound with respect to the skeleton-based
   approximation, then, at the next iteration, this property still holds. (A
   fortiori, this also works for a type-based pre-fixpoint. We use the
   (somewhat trivial) facts that the skeleton-based approximation is natural
   and nice. In order to deal with the bullet cases, we perform induction over
   the tail-finiteness of [ty1] and [ty2]. There is no control over the number
   of times this loop might run: this explains why we must have a fixed point
   (as opposed to an over-approximation) on the left-hand side. *)

Lemma basic_rsoundness:
  forall selfty selfsk,
  rsound selfty selfsk ->
  natural selfsk ->
  nice selfsk ->
  pointwise_inclusion selfty (subtype selfty) ->
  rsound (subtype selfty) (subskeleton selfsk).
Proof.
  intros ? ? h na ni prefixpoint ? ? ? ?.
  assert (tfty1 : tf ty1); eauto.
  assert (tfty2 : tf ty2); eauto.
  induction tfty1 as [ | ty11 ty12 | ty1 ].
  (* Leaf *)
    induction tfty2 as [ | | ty2 ]; intros k hk; simpl in hk; try solve [ false; auto ].
    (* Leaf Leaf *)
    rewrite sk_leaf. rewrite sk_leaf. simpl. auto.
    (* Leaf Bullet *)
    rewrite sk_bullet.
    rewrite (nice_subskeleton _ ni)
      with (t1 := shift 0 (sk 0 TyLeaf fty1))
           (t2 := shift 1 (sk 0 ty2 (finite_hereditary_bullet (TyBullet ty2) fty2 ty2 refl)));
    [| eauto using shift_zero | eauto using sk_shift_zero ].
    rewrite (natural_subskeleton _ na).
    replace (Z_of_nat 0 - Z_of_nat 1) with (-1); try solve [ simpl; auto ].
    eapply back_to_set_inclusion. eassumption.
    apply translate_monotonic; eauto using set_inclusion_transitive. (* prefixpoint and IHtfty2 are used *)
  (* Arrow *)
    clear tfty1 IHtfty1. (* required for the following induction to have the right shape *)
    induction tfty2 as [ | ty21 ty22 | ty2 ]; intros k hk; simpl in hk; try solve [ false; auto ].
    (* Arrow Arrow *)
    rewrite sk_arrow. rewrite sk_arrow. simpl.
    eapply back_to_set_inclusion. eassumption.
    apply combine_monotonic; eauto.
    (* Arrow Bullet *)
    rewrite sk_bullet.
    rewrite (nice_subskeleton _ ni)
      with (t1 := shift 0 (sk 0 (TyArrow ty11 ty12) fty1))
           (t2 := shift 1 (sk 0 ty2 (finite_hereditary_bullet (TyBullet ty2) fty2 ty2 refl)));
    [| eauto using shift_zero | eauto using sk_shift_zero ].
    rewrite (natural_subskeleton _ na).
    replace (Z_of_nat 0 - Z_of_nat 1) with (-1); try solve [ simpl; auto ].
    eapply back_to_set_inclusion. eassumption.
    apply translate_monotonic; eauto using set_inclusion_transitive. (* prefixpoint and IHtfty2 are used *)
  (* Bullet *)
  rewrite sk_bullet.
  rewrite (nice_subskeleton _ ni)
    with (t1 := shift 1 (sk 0 ty1 (finite_hereditary_bullet (TyBullet ty1) fty1 ty1 refl)))
         (t2 := shift 0 (sk 0 ty2 fty2));
  [| eauto using sk_shift_zero | eauto using shift_zero ].
  rewrite (natural_subskeleton _ na).
  replace (Z_of_nat 1 - Z_of_nat 0) with 1; try solve [ simpl; auto ].
  replace (subtype selfty (TyBullet ty1) ty2) with (translate 1 (selfty ty1 ty2)); try solve [ simpl; auto ].
  apply translate_monotonic.
  eauto using set_inclusion_transitive. (* prefixpoint and IHtfty1 are used *)
Qed.

(* The above soundness lemma is slightly too basic: it does imply that any
   pre-fixpoint of the type-based algorithm is sound, but does not imply that
   the sequence of type-based approximations reaches a limit. In order to
   prove that the sequence converges, we need a slightly more precise result:
   if, at some point, the type-based approximation is sound with respect to
   the skeleton-based approximation, then, at some later point, the type-based
   approximation will be sound with respect to the next skeleton-based
   approximation. This relaxed statement is required because the type-based
   sequence converges more slowly than the skeleton-based sequence. *)

Fixpoint iterate (A : Type) (f : A -> A) (n : nat) (x : A) : A :=
  match n with
  | 0%nat =>
      x
  | S n =>
      f (iterate A f n x)
  end.

Implicit Arguments iterate [ A ].

Lemma iterate_rev:
  forall A (f : A -> A) n x,
  f (iterate f n x) = iterate f n (f x).
Proof.
  induction n; simpl; auto. intro. f_equal. eauto.
Qed.

Definition prsound (selfty : ty -> ty -> set) (selfsk : skeleton -> skeleton -> set) : Prop :=
  forall ty1 ty2 fty1 fty2,
  exists n,
  set_inclusion
    (iterate subtype n selfty ty1 ty2)
    (selfsk (sk 0 ty1 fty1) (sk 0 ty2 fty2)).

Require Import Coq.Arith.Max.

Lemma decreasing_selfty_preliminary:
  forall selfty,
  pointwise_inclusion (subtype selfty) selfty ->
  forall n j,
  pointwise_inclusion (iterate subtype (j + n) selfty) (iterate subtype n selfty).
Proof.
  intros ? ? ?. induction j; simpl. eauto.
  apply pointwise_inclusion_transitive with (f2 := iterate subtype (S n) selfty).
  apply subtype_monotonic. eauto. simpl.
  clear j IHj. induction n; simpl; eauto using subtype_monotonic.
Qed.

Lemma decreasing_selfty:
  forall selfty,
  pointwise_inclusion (subtype selfty) selfty ->
  forall n1 n2,
  (n1 <= n2)%nat ->
  pointwise_inclusion (iterate subtype n2 selfty) (iterate subtype n1 selfty).
Proof.
  intros. replace n2 with ((n2 - n1) + n1)%nat; try omega. eauto using decreasing_selfty_preliminary.
Qed.

Lemma decreasing_selfty_corollary:
  forall selfty,
  pointwise_inclusion (subtype selfty) selfty ->
  forall n1 n2,
  (n1 <= n2)%nat ->
  forall ty1 ty2 s,
  set_inclusion (iterate subtype n1 selfty ty1 ty2) s ->
  set_inclusion (iterate subtype n2 selfty ty1 ty2) s.
Proof.
  introv h1 h2 h3.
  generalize (decreasing_selfty _ h1 _ _ h2 ty1 ty2); intro.
  eauto using set_inclusion_transitive.
Qed.

Lemma precise_rsoundness2:
  forall selfty selfsk,
  prsound selfty selfsk ->
  natural selfsk ->
  nice selfsk ->
  pointwise_inclusion (subtype selfty) selfty ->
  prsound selfty (subskeleton selfsk).
Proof.
  intros ? ? h na ni postfixpoint ? ? ? ?.
  assert (tfty1 : tf ty1); eauto.
  assert (tfty2 : tf ty2); eauto.
  induction tfty1 as [ | ty11 ty12 | ty1 ].
  (* Leaf *)
    induction tfty2 as [ | | ty2 ].
    (* Leaf Leaf *)
    exists 1%nat. intros k hk; simpl in hk.
    rewrite sk_leaf. rewrite sk_leaf. simpl. auto.
    (* Leaf Arrow *)
    exists 1%nat. rewrite sk_leaf. rewrite sk_arrow. simpl. auto.
    (* Leaf Bullet *)
    generalize (IHtfty2 (finite_hereditary_bullet (TyBullet ty2) fty2 ty2 refl)); intros [ n IH ].
    exists (S n). intros k hk; simpl in hk.
    rewrite sk_bullet.
    rewrite (nice_subskeleton _ ni)
      with (t1 := shift 0 (sk 0 TyLeaf fty1))
           (t2 := shift 1 (sk 0 ty2 (finite_hereditary_bullet (TyBullet ty2) fty2 ty2 refl)));
    [| eauto using shift_zero | eauto using sk_shift_zero ].
    rewrite (natural_subskeleton _ na).
    replace (Z_of_nat 0 - Z_of_nat 1) with (-1); try solve [ simpl; auto ].
    eapply back_to_set_inclusion. eassumption.
    eauto. (* IH is used *)
  (* Arrow *)
    clear tfty1 IHtfty1. (* required for the following induction to have the right shape *)
    induction tfty2 as [ | ty21 ty22 | ty2 ].
    (* Arrow Leaf *)
    exists 1%nat. rewrite sk_leaf. rewrite sk_arrow. simpl. auto.
    (* Arrow Arrow *)
    generalize (h ty21 ty11
      (finite_hereditary_domain (TyArrow ty21 ty22) fty2 ty21 ty22 refl)
      (finite_hereditary_domain (TyArrow ty11 ty12) fty1 ty11 ty12 refl)
    ); intros [ n1 h1 ].
    generalize (h ty12 ty22
      (finite_hereditary_codomain (TyArrow ty11 ty12) fty1 ty11 ty12 refl)
      (finite_hereditary_codomain (TyArrow ty21 ty22) fty2 ty21 ty22 refl)
    ); intros [ n2 h2 ].
    exists (S (max n1 n2)).
    intros k hk; simpl in hk.
    rewrite sk_arrow. rewrite sk_arrow. simpl.
    eapply back_to_set_inclusion. eassumption.
    apply combine_monotonic.
    apply decreasing_selfty_corollary with (n1 := n1). assumption. apply le_max_l. eauto.
    apply decreasing_selfty_corollary with (n1 := n2). assumption. apply le_max_r. eauto.
    (* Arrow Bullet *)
    generalize (IHtfty2 (finite_hereditary_bullet (TyBullet ty2) fty2 ty2 refl)); intros [ n IH ].
    exists (S n). intros k hk; simpl in hk.
    rewrite sk_bullet.
    rewrite (nice_subskeleton _ ni)
      with (t1 := shift 0 (sk 0 (TyArrow ty11 ty12) fty1))
           (t2 := shift 1 (sk 0 ty2 (finite_hereditary_bullet (TyBullet ty2) fty2 ty2 refl)));
    [| eauto using shift_zero | eauto using sk_shift_zero ].
    rewrite (natural_subskeleton _ na).
    replace (Z_of_nat 0 - Z_of_nat 1) with (-1); try solve [ simpl; auto ].
    eapply back_to_set_inclusion. eassumption. eauto. (* IH is used *)
  (* Bullet *)
  generalize (IHtfty1 (finite_hereditary_bullet (TyBullet ty1) fty1 ty1 refl)); intros [ n IH ].
  exists (S n).
  rewrite sk_bullet.
  rewrite (nice_subskeleton _ ni)
    with (t1 := shift 1 (sk 0 ty1 (finite_hereditary_bullet (TyBullet ty1) fty1 ty1 refl)))
         (t2 := shift 0 (sk 0 ty2 fty2));
  [| eauto using sk_shift_zero | eauto using shift_zero ].
  rewrite (natural_subskeleton _ na).
  simpl. eauto. (* wow, that was unexpected! *)
Qed.

Lemma prsound_below:
  forall selfty selfsk,
  prsound selfty selfsk ->
  pointwise_inclusion (subtype selfty) selfty ->
  prsound (subtype selfty) selfsk.
Proof.
  intros ? ? h ? ? ? ? ?.
  generalize (h ty1 ty2 fty1 fty2); intros [ n f ].
  exists n.
  apply set_inclusion_transitive with (s2 := iterate subtype n selfty ty1 ty2); eauto.
  generalize ty1 ty2. clear ty1 ty2 fty1 fty2 f.
  induction n; simpl. eauto. eapply subtype_monotonic. eauto.
Qed.

(* To summarize the above results, we relate the two sequences of
   approximations. The relation is a simulation where the left-hand side (the
   sequence of type-based approximations) *)

CoInductive simulation : sequence (ty -> ty -> set) -> sequence (skeleton -> skeleton -> set) -> Prop :=
| Simulation :
  forall ts ss selfty selfsk,
  nice selfsk ->
  natural selfsk ->
  pointwise_inclusion (subskeleton selfsk) selfsk ->
  pointwise_inclusion (subtype selfty) selfty ->
  prsound selfty selfsk ->
  rcomplete selfty selfsk ->
  simulation ts ss ->
  simulation (SeqCons ts selfty) (SeqCons ss selfsk).

Lemma exhibit_simulation_preliminary:
  forall selfty selfsk,
  nice selfsk ->
  natural selfsk ->
  pointwise_inclusion (subskeleton selfsk) selfsk ->
  pointwise_inclusion (subtype selfty) selfty ->
  prsound selfty selfsk ->
  rcomplete selfty selfsk ->
  simulation (orbit subtype selfty) (orbit subskeleton selfsk).
Proof.
  cofix.
  intros.
  rewrite (seq_decompose (orbit subtype _)); simpl.
  rewrite (seq_decompose (orbit subskeleton _)); simpl.
  econstructor; eauto.
  eapply exhibit_simulation_preliminary; eauto.
  eapply nice_subskeleton; eauto.
  eapply natural_subskeleton; eauto.
  eapply subskeleton_monotonic. eauto.
  eapply subtype_monotonic. eauto.
  eapply precise_rsoundness2; eauto.
  eapply prsound_below; eauto.
  eapply subtype_monotonic. eauto.
  eapply basic_rcompleteness; eauto.
Qed.

Theorem exhibit_simulation:
  simulation (orbit subtype (@top ty)) (orbit subskeleton (@top skeleton)).
Proof.
  apply exhibit_simulation_preliminary.
  apply nice_top.
  apply natural_top.
  intros ? ?. intuition eauto.
  intros ? ?. intuition eauto.
  intros ? ? ? ?. exists 0%nat. unfold top. simpl. auto.
  intros ? ? ? ?. unfold top. simpl. auto.
Qed.

(* TEMPORARY je suis bete, j'ai deja demontre que pour ty1, ty2 fixes la
   sequence de gauche converge! autrement dit, le passage par une simulation
   precise est overkill *)

Lemma limit_bar:
  forall limit ty1 ty2 selfty,
  always (now (= limit)) (sample ty1 ty2 (orbit subtype selfty)) ->
  limit = selfty ty1 ty2.
Proof.
  introv a.
  rewrite (seq_decompose (sample _ _ _)) in a; simpl in a.
  generalize (A_cons_head a). auto.
Qed.

Lemma iterate_limit_preliminary:
  forall limit ty1 ty2 selfty,
  always (now (= limit)) (sample ty1 ty2 (orbit subtype selfty)) ->
  subtype selfty ty1 ty2 = selfty ty1 ty2.
Proof.
  introv a.
  rewrite (limit_bar limit ty1 ty2 selfty) in a; auto.
  rewrite (seq_decompose (sample _ _ _)) in a; simpl in a.
  generalize (A_cons_tail a).
  rewrite (seq_decompose (map _ _)); simpl. intro b.
  generalize (A_cons_head b). simpl. auto.
Qed.

Lemma iterate_limit:
  forall limit ty1 ty2 n selfty,
  always (now (= limit)) (sample ty1 ty2 (orbit subtype selfty)) ->
  iterate subtype n selfty ty1 ty2 = selfty ty1 ty2.
Proof.
  induction n; simpl; intros ? a.
  auto.
  rewrite iterate_rev. rewrite <- (iterate_limit_preliminary _ _ _ _ a). apply IHn.
  rewrite (seq_decompose (sample _ _ _)) in a; simpl in a.
  generalize (A_cons_tail a). auto.
Qed.

Lemma iterate_limit_foo:
  forall limit ty1 ty2 fty1 fty2 selfty selfsk,
  always (now (= limit)) (sample ty1 ty2 (orbit subtype selfty)) ->
  prsound selfty selfsk ->
  rcomplete selfty selfsk ->
  selfty ty1 ty2 = selfsk (sk 0 ty1 fty1) (sk 0 ty2 fty2).
Proof.
  intros ? ? ? ? ? ? ? a s c.
  generalize (s ty1 ty2 fty1 fty2); intros [ n ss ].
  generalize (c ty1 ty2 fty1 fty2); intro.
  rewrite (iterate_limit _ _ _ n _ a) in ss.
  eauto.
Qed.

Lemma convergence:
  forall limit ty1 ty2 fty1 fty2 selfty selfsk,
  simulation (orbit subtype selfty) (orbit subskeleton selfsk) ->
  always (now (= limit)) (sample ty1 ty2 (orbit subtype selfty)) ->
  always (now (= limit)) (sample (sk 0 ty1 fty1) (sk 0 ty2 fty2) (orbit subskeleton selfsk)).
Proof.
  intros ? ? ? ? ?. cofix. intros selfty selfsk sim a.
  rewrite (seq_decompose (orbit subtype _)) in sim;
  rewrite (seq_decompose (orbit subskeleton _)) in sim;
  simpl in sim.
  dependent destruction sim. 
  rewrite (seq_decompose (sample _ _ _)); simpl. 
  constructor.
  eapply convergence. eassumption.
  rewrite (seq_decompose (sample _ _ _)) in a; simpl in a. unfold sample. eapply A_cons_tail. eassumption.
  unfold now.
  rewrite (limit_bar limit ty1 ty2 selfty a); eauto.
  rewrite (iterate_limit_foo limit ty1 ty2 fty1 fty2 selfty selfsk a); eauto.
Qed.

Lemma babi:
  forall limit ty1 ty2 fty1 fty2,
  forall s,
  eventually (always (now (= limit))) s ->
  forall selfty selfsk,
  s = sample ty1 ty2 (orbit subtype selfty) ->
  simulation (orbit subtype selfty) (orbit subskeleton selfsk) ->
  eventually (always (now (= limit))) (sample (sk 0 ty1 fty1) (sk 0 ty2 fty2) (orbit subskeleton selfsk)).
Proof.
  induction 1; introv ? sim.
  apply EventuallyNow. eapply convergence. eassumption. congruence.
  rewrite (seq_decompose (sample _ _ _)); simpl.
  apply EventuallyLater. apply IHeventually with (selfty := subtype selfty).
  rewrite (seq_decompose (sample _ _ _)) in H0; simpl in H0. unfold sample.
  congruence.
  rewrite (seq_decompose (orbit subtype _)) in sim;
  rewrite (seq_decompose (orbit subskeleton _)) in sim;
  simpl in sim.
  dependent destruction sim.
  auto. 
Qed.

Lemma convergence2:
  forall limit ty1 ty2 fty1 fty2 selfty selfsk,
  simulation (orbit subtype selfty) (orbit subskeleton selfsk) ->
  eventually (always (now (= limit))) (sample ty1 ty2 (orbit subtype selfty)) ->
  eventually (always (now (= limit))) (sample (sk 0 ty1 fty1) (sk 0 ty2 fty2) (orbit subskeleton selfsk)).
Proof.
  eauto using babi.
Qed.

Theorem coincidence:
  forall ty1 ty2 fty1 fty2,
  exists limit,
  eventually (always (now (= limit))) (approximations ty1 ty2) /\
  eventually (always (now (= limit))) (sample (sk 0 ty1 fty1) (sk 0 ty2 fty2) (orbit subskeleton (@top skeleton))).
Proof.
  intros.
  generalize (approximations_converges ty1 (finite_tf _ fty1) ty2 (finite_tf _ fty2)). intros [ limit ea ].
  exists limit. split.
  assumption.
  apply convergence2 with (selfty := @top ty).
  apply exhibit_simulation. auto.
Qed.

(* TEMPORARY pour terminer, montrer que cette limite est bien l'ensemble
   attendu -- i.e. il est sound et complet vis-`a-vis de sksub *)

